# To generate a .dotfile from the gamergate data files: 
# python dotgenerator.py 
# To query the files for a subject:
# python dotgenerator.py "Zoe Quinn" 

import re
import sys
import os
import glob

# You may need to install these libraries first using pip or easy_install
import pandas as pd # for reading excel files
import xlrd # for pandas
# import daemon # from python-daemonize 

import BaseHTTPServer

# Debugging 
import time # for timing a function
import cProfile # for profiling the program 




## Convert a wiki-style .txt file to a zq_notes style file.
def dewikify(filename, reader):
    regex = re.compile(r"(=+)([^=]+)(=+)")
    with open(filename, "r") as f:
        depth_stack = list([0])
        word_stack = list([reader.title_from_filepath(filename)])
        last_word = ""
        last_depth = 0
        for line in f:

            # Strip wikilink symbols for easier reading
            line = re.sub("[\[\]]", "", line)
        
            matches = regex.match(line)
            if(matches is None):
                yield line
            else:

                new_depth = len(matches.group(1))
                new_word = matches.group(2).strip()

                while(len(depth_stack) > 0 and new_depth <= last_depth):
                    last_depth = depth_stack.pop()
                    last_word = word_stack.pop()

                depth_stack.append(last_depth)
                word_stack.append(last_word)

                depth_stack.append(new_depth)
                word_stack.append(new_word)

                word = new_word
                if len(last_word):
                    word = "%s/%s" % (last_word,new_word)

                yield "------\n" # section break for text file format
                yield "%s\n" % (word)

                last_word = word
                last_depth = new_depth
    f.close() 




## Convert a chunk of text into (title,body) 
class ChunkReader:
    MODE_SEEKTITLE = 0   
    MODE_SEEKDATA = 1   
    MODE_READDATA = 2
    mode = MODE_SEEKTITLE
    title = ""
    data = ""
    buf = "" # avoid saving trailing WS in data

    def __init__(self): 
        self.reset()

    def reset(self): 
        self.mode = self.MODE_SEEKTITLE
        self.title = ""
        self.data = ""
        self.buf = "" 

    def feedLine(self, line): 
        line_stripped = line.strip()

        # Skip blank lines unless you are already reading data.
        if(len(line_stripped) == 0): 
            if(self.mode == self.MODE_READDATA):
                self.buf = "%s\n" % (self.buf)

        # First nonblank line is the title.
        elif(self.mode==self.MODE_SEEKTITLE):
            self.title = line_stripped
            self.mode=self.MODE_SEEKDATA
        # Next nonblank line begins data
        else:
            if(self.mode==self.MODE_SEEKDATA):
                self.mode=self.MODE_READDATA
            if(self.mode==self.MODE_READDATA):
                self.data = "%s%s%s" % (self.data, self.buf, line)
                self.buf = ""

    def getTuple(self):
        return [self.title, self.data]


## Data and metadata to be stored 
class DataObject: 
    # Data
    title=""
    data=""

    def __init__(self, title, data, filename="", section="", line=-1, byte=-1):
        self.title = title
        self.data = data
        self.fn = filename
        self.section = section
        self.line = line # Not assigned
        self.byte = byte # Not assigned

        self.fulltext = "" # Cache for repeated access

    ## Output formatting 
    def textDecorateFilename(self):
        result =""
        if(len(self.fn)):
            result= "origin: %s" % (self.fn)
        return result

    def textDecorateSection(self):
        result =""
        if(len(self.fn)):
            result= "section: %s" % (self.fn)
        return result

    def textDecorateLineNumber(self):
        result =""
        if(self.line >= 0):
            result= "line: %s" % (self.line)
        return result

    def textDecorateByteNumber(self):
        result =""
        if(self.byte >= 0):
            result= "byte: %s" % (self.byte)
        return result

    def textDecorationsJSON(self):
        text = "{"
        results = [
                self.textDecorateFilename(),
                self.textDecorateSection(),
                self.textDecorateLineNumber(),
                self.textDecorateByteNumber(),
        ]
        for r in results:
            if(len(r) > 0):
                if(len(text) > 1):
                    text = "%s, %s" % (text, r)
                else:
                    text = "%s%s" % (text, r)
        text = "%s}" % (text)
        return text 

    def getText(self):
        if len(self.fulltext) == 0:
            origin = self.textDecorationsJSON()
            origin = "" # Squelch 
            self.fulltext = "%s\n%s\n%s\n" % (self.title, origin, self.data)
        return self.fulltext 


# Store multiple results under one title
class DataCollectionManager:
    title=""
    objects=[] 
    def __init__(self, title, do=None):
        self.title = title
        self.objects = []

        if not(do==None):
            self.objects = [do]

    def getText(self):
        s = "" 
        for obj in self.objects:
            s = "%s%s" % (s, obj.getText())
        return s


# WordSearcher speeds up the text processing by reducing the number of
# times that the text is scanned. 
# Instead of regexing each text for each title, we read each text word
# for word and check for the title's presence when any word matches the
# first word of the title.
# Time with old algorithm: 
# 4973.706u 6.991s 1:23:19.10 99.6%       5+329k 3+2io 0pf+0w
# Time using WordSearcher:
# 127.321u 7.597s 2:15.32 99.6%   5+327k 0+2io 0pf+0w
class WordSearcher:
    titleFirstWords = dict([])

    def __init__(self, r): # r:MyReader
        reflags = re.IGNORECASE | re.MULTILINE | re.UNICODE
        self.re_words = re.compile(r"(\S+)", reflags)

        self.build_index(r)

    # Save all titles to a hash whose key is the first word in the title
    def build_index(self, r):
        self.titleFirstWords = dict([])

        for key in r.hashData:

            matches = self.re_words.match(key)

            if(matches): 
                firstword = matches.group(1)
                if not(firstword in self.titleFirstWords):
                    self.titleFirstWords[firstword] = []

                self.titleFirstWords[firstword].append(key)
            else:
                print "DEBUG: THIS SHOULD NEVER HAPPEN"
                print "...key:%s" % (key)


        # Sort the titles by length, from largest to smallest.
        # This allows the searcher to choose only the largest title.
        for key in self.titleFirstWords:
            self.titleFirstWords[key].sort(lambda x,y: cmp(len(y), len(x)))

    def searchDataObject(self, do):

        results = []
        results_key = dict()
        s = do.getText()

        for w in self.re_words.finditer(s):
            start = w.start()

            firstword_match = self.re_words.match(s, start)
            firstword = firstword_match.group(1)

            if(firstword in self.titleFirstWords):
                for title in self.titleFirstWords[firstword]:

                    # See if we already link to the target
                    if title in results_key:
                        continue
                    
                    # Do not link back to yourself
                    if title == do.title:
                        continue

                    match = re.compile(title).match(s, start)
                    if(match):
                        results.append(title)
                        results_key[title] = 1 

                        # Take only the longest title match at this point.
                        # Remove this break if you want every keyword match.
                        break

        return results

    # A test that only scans words for testing the timing
    def searchDataObject_noop(self, do):

        results = []
        results_key = dict()
        s = do.getText()

        for w in self.re_words.finditer(s):
            start = w.start()

            firstword_match = self.re_words.match(s, start)
            firstword = firstword_match.group(1)

        return results


class MyReader:
    hashData = dict([])
    hashReferences = dict([])
    warnings=[]

    def __init__(self, inputFiles):
        self.hashData = dict([])
        self.hashReferences = dict([])
        self.warnings = []

        # Read all input files
        import os.path
        for fn in inputFiles:
            if not (os.path.isfile(fn)): 
                self.warn("file not found: %s" % (fn))
                continue 

            # Select the reader function
            f = self.parse_file_dashbreak
            if (".xls" in fn):
                f = self.parse_file_excel
            elif (".csv" in fn):
                f = self.parse_file_flat

            f(fn)


    def warn(self, message):
        self.warnings.append(message)
        print "Warning: %s" % (message) 

    def query(self, term, request="ref", reflags=re.IGNORECASE | re.MULTILINE | re.UNICODE):
        # NOTE: \b will fail to match terms that begin or end with a symbol. 
        # The ending period is for the form "see also Foo (bar)."
        query = r'(^|[\b\s])%s([\b\s\.,]|$)' % (re.escape(term))
        keymatcher = re.compile(query, reflags)
        
        for key in self.hashData:
            text = self.hashData[key].getText()
            if(keymatcher.search(text)):
                if(request == "text"):
                    yield text
                elif(request == "obj"):
                    yield self.hashData[key]
                else: 
                    yield key


    # Query multiple terms (not tested, not used)
    def query_multi(self, terms, request="ref", reflags=re.IGNORECASE | re.MULTILINE | re.UNICODE):
        i = 0 # counter 
        results = dict([])
        queries = dict([])
        for term in terms:
            results[term] = [] 
            query = r'(^|[\b\s])%s([\b\s\.,]|$)' % (re.escape(term))
            queries[term] = re.compile(query, reflags)

        for key in self.hashData:
            text = self.hashData[key].getText()

            for term in terms: 
                if key == term:
                    continue

                if(queries[term].search(text)):

                    if(request == "text"):
                        results[term].append(text)
                    else: 
                        results[term].append(key)

            # display a counter 
            i = i + 1
            if(i % 100 == 0):
                print "%d / %d" % (i, len(self.hashData))

        return results


    # Generate the edges between subjects 
    def buildReferences(self):

        # Build the WordSearcher index from our hash's keys
        ws = WordSearcher(self)

        print "Building references for %d records" % (len(self.hashData))
        i = 0
        for key in self.hashData:

            self.hashReferences[key] = ws.searchDataObject(self.hashData[key])

            # Display a counter
            i = i + 1
            if(i % 100 == 0):
                print "%d / %d" % (i, len(self.hashData))


    # Old version of buildReferences that took an hour to run 3000 records.
    def buildReferences_single(self):

        records_total = len(self.hashData)
        records_current = 0

        print "Building references for %d records" % (records_total)
        sys.stdout.flush()

        keys_seen = dict()

        for outerkey in self.hashData:

            self.hashReferences[outerkey] = list()
            matches = self.query(outerkey)

            try:
                matches.remove(outerkey)
            except ValueError, e: 
                print "Warning: key:%s not in set:%s" % (outerkey, matches)

            self.hashReferences[outerkey] = matches 

            # update display counter
            records_current += 1
            if(records_current % 100 == 0):
                print "%d / %d" % (records_current, records_total)
                sys.stdout.flush()


    def getReferences(self):
        for key in self.hashReferences:
            for ref in self.hashReferences[key]:
                yield (key, ref)

    def getDotFileLines(self):
        f_escaper = lambda x: x.replace('"', '\\"')
        yield "digraph A {"
        for key, ref in self.getReferences():
            yield "\"%s\" -> \"%s\"" % ( f_escaper(ref), f_escaper(key))
        yield "}"


    ### Input functions ###

    def data_save(self, dataObject):
        key = dataObject.title 

        if not(key in self.hashData):
            self.hashData[key] = DataCollectionManager(key)

        self.hashData[key].objects.append(dataObject)

    def title_from_filepath(self, filename):
        return os.path.splitext(os.path.basename(filename))[0].title()

    def parse_file_flat(self, filename):
        f = open(filename, "r")
        data = f.read()
        title = self.title_from_filepath(filename)
        self.data_save(DataObject(title, data, filename, title.lower()))

    def parse_file_excel(self, filename):

        title=self.title_from_filepath(filename)

        xl = pd.ExcelFile(filename)
        for sheetID in xl.sheet_names:
            df = xl.parse(sheetID)
            data = df.to_csv(encoding="utf-8")
            self.data_save(DataObject(title, data, filename, sheetID))

    def parse_file_dashbreak(self, filename):
        cr = ChunkReader()


        ### HACK ### 
        fstream = None
        wiki_styles = [
    "../censorship/huffpo.txt",
    "../censorship/reddit.txt",
    "../censorship/twitch.txt",
    "../censorship/twitter.txt",
    "../censorship/wikipedia.txt",
    "../censorship/wikipedia_and_gamergate.txt",
    "../censorship/wikipedia_controversies.txt",
    "../censorship/wikipedia_subjects.txt"
        ]
        if(filename in wiki_styles):

            fstream = dewikify(filename,self)
        else:
            fstream = open(filename)
        ### /end HACK ### 


        line_no = 0
        byte_no = 0
        section_no = 0
        section_line_no = 0
        section_byte_no = 0

        section_break = re.compile("^-----")


        flag_espen = 0

        flag_new_section = 1
        while(flag_new_section):
            flag_new_section = 0 
            section_bytes = 0
            section_lines = 0

            for line in fstream:
                section_bytes += len(line) # Probably a BUG not counting \n
                section_lines += 1

                if(section_break.match(line)):
                    flag_new_section=1 # Let outer loop continue
                    break

                cr.feedLine(line) 

            # loop broke on either new section or EOF 
            (title, data) = cr.getTuple()
            if(len(title)): 
                d_o = DataObject(title, data, filename, section_no, section_line_no, section_byte_no)
                self.data_save(d_o)

            section_no += 1
            section_line_no = line_no
            section_byte_no = byte_no

            line_no = line_no + section_lines
            byte_no = byte_no + section_bytes

            cr.reset()

        fstream.close()

class MyRunner:

    @classmethod
    def subdirs_exist(cls, path, subdirs):
        for test in subdirs:
            if not (os.path.isdir("%s%s" %(path,test))):
                return False
        return True

    @classmethod
    def buildFilePath(cls):

        testpaths = ["../", "../../"]
        subdirs=["subjects","data"]

        for path in testpaths: 
            if(MyRunner.subdirs_exist(path, subdirs)):
                return path

        return ""

    @classmethod
    def run(cls):
        filepath=MyRunner.buildFilePath()
        datasources = [ 
    "zq_notes.txt",
    "wikia_export.txt",
    "wikia_manual_updates.txt",
    "natsec/alavi.txt",
    "natsec/bilderberg.txt",
    "natsec/doha2017.txt",
    "natsec/egypt.txt",
    "natsec/ffeu.txt",
    "natsec/fortuyn.txt",
    "natsec/goldenchain.txt",
    "natsec/hna_group.txt",
    "natsec/icg.txt",
    "natsec/isd.txt",
    "natsec/isgp-libcia.txt",
    "natsec/jstreet.txt",
    "natsec/gazamarch.txt",
    "natsec/goldenchain.txt",
    "natsec/madrid.txt",
    "natsec/natsec.txt",
    "natsec/niac.txt",
    "natsec/oxfam.txt",
    "natsec/pol_hli.txt",
    "natsec/qatar.txt",
    "natsec/uae.txt",
    "natsec/wef.txt",
    "natsec/worldfuture.txt",
    "subjects/300committee.txt",
    "subjects/adl.txt",
    "subjects/adler.txt",
    "subjects/atheismplus.txt",
    "subjects/bmcdig.txt",
    "subjects/capital_guidance.txt",
    "subjects/demshadowparty.txt",
    "subjects/esa.txt",
    "subjects/hava.txt",
    "subjects/hcfactions.txt",
    "subjects/newsom_benello.txt",
    "subjects/opgamergate.txt",
    "subjects/pedowood.txt",
    "subjects/splc.txt",
    "subjects/sonyleaks.txt",
    "subjects/sydette.txt",
    "data/avaaz.xls",
    "data/conleaks.xls",
    "data/freethefive.csv",
    "data/gamersaredead.xls",
    "data/gamedrops.csv",
    "data/hackers.xls",
    "data/ludology.xls",
    "data/philfish.xls",
    "data/swedes.csv",
    "censorship/4chan.txt",
    "censorship/facebook.txt",
    "censorship/googleplay.txt",
    "censorship/hackernews.txt",
    "censorship/huffpo.txt",
    "censorship/imgur.txt",
    "censorship/instagram.txt",
    "censorship/reddit.txt",
    "censorship/twitch.txt",
    "censorship/twitter.txt",
    "censorship/wikipedia.txt",
    "censorship/wikipedia_and_gamergate.txt",
    "censorship/wikipedia_controversies.txt",
    "censorship/wikipedia_subjects.txt",
    "censorship/youtube.txt",

    #"offtopic/belzian.txt",
    #"offtopic/pizzagate.txt",
]

        # HACK: Speed up loading times greatly 
        # datasources = [ "zq_notes.txt" ]
        # datasources = ["natsec/fortuyn.txt" ]
        #datasources = [ "natsec/jewsarsour.txt" ]
        datasources = map( lambda x: "%s%s" % (filepath, x), datasources)

        return MyReader(datasources)



def run_dotfile(r): 
    DOTFILE= open("out.dot", "w")
    r.buildReferences()
    for line in r.getDotFileLines():
        DOTFILE.write("%s\n" % line) 

def run_query(r, txt):
    results = list(r.query(txt, "text"))
    if len(results):
        print results[0]
        results = results[1:]
    for item in results:
        print "\n-----\n"
        print item


def timedfunction(f, runstr=""):
    t1 = time.clock()

    print "Running %s..." % (runstr)
    result = f()

    t2 = time.clock()

    print "Time elapsed: %f" % (t2)

    return result



def run(): 

    r = MyRunner.run() 
    # r = timedfunction(MyRunner.run)

    if len(sys.argv) > 1:
        run_query(r, sys.argv[1])
    else:
        run_dotfile(r)


if __name__ == '__main__':
    run()
    #cProfile.run( 'run()' ) 
